# Generated by Django 3.1.2 on 2020-10-15 05:21

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="CsvDownload",
            options={"verbose_name": "CSV Download"},
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("filename", models.CharField(max_length=100)),
                (
                    "timestamp",
                    models.DateTimeField(
                        auto_now_add=True, help_text="When the download took place."
                    ),
                ),
                (
                    "row_count",
                    models.IntegerField(
                        blank=True, help_text="Rows downloaded", null=True
                    ),
                ),
                (
                    "columns",
                    models.CharField(
                        help_text="The list of source columns included in the download",
                        max_length=500,
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="User who initiated the download.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



"""Optional functions for uploading data direct to S3."""
import contextlib
from io import BytesIO, TextIOWrapper
from tempfile import TemporaryFile
from typing import IO, Generator, Tuple, Union

from django.db.models import QuerySet

from .csv import write_csv
from .settings import MAX_ROWS

try:
    import boto3
except ImportError:
    raise ImportError("You cannot use the django_csv.s3 module without boto3.")

# type used to smooth over TemporaryFile <> BytesIO mismatch
FileLikeObject = Union[IO[bytes], BytesIO]
# type to represent an S3 address (bucket, key)
S3Url = Tuple[str, str]


# extracted out to facilitate testing
def _put_object(bucket: str, key: str, fileobj: FileLikeObject) -> None:
    """Upload binary stream to S3 using put_pubject."""
    client = boto3.client("s3")
    client.put_object(Bucket=bucket, Key=key, Body=fileobj, ContentType="text/csv")


# extracted out to facilitate testing
def _upload_fileobj(bucket: str, key: str, fileobj: FileLikeObject) -> None:
    """Upload binary stream to S3 using upload_fileobj."""
    client = boto3.client("s3")
    client.upload_fileobj(fileobj, bucket, key)


@contextlib.contextmanager
def s3_upload_multipart(bucket: str, key: str) -> Generator:
    """
    Context manager used to write to S3 using upload_fileobj.

    This context manager writes to a TemporaryFile and then uses the
    multipart boto3 upload function `upload_fileobj`, and is more
    appropriate for large files.

    >>> with s3_upload_fileobj("bucket", "obj_key") as fileobj:
    ...     write_csv(fileobj, queryset, "col1", "col2")

    """
    with TemporaryFile() as fileobj:
        with TextIOWrapper(
            fileobj, encoding="utf-8", newline="", write_through=True
        ) as buffer:
            yield buffer
            fileobj.seek(0)
            _upload_fileobj(bucket, key, fileobj)


@contextlib.contextmanager
def s3_upload(bucket: str, key: str) -> Generator:
    """
    Context manager used to write to S3 using put_object.

    This context manager writes to a in-memory buffer and then uses the
    one-shot boto3 upload function `put_object`, and is more appropriate
    for smaller files.

    >>> with s3_put_object("bucket", "obj_key") as fileobj:
    ...     write_csv(fileobj, queryset, "col1", "col2")

    """
    with BytesIO() as fileobj:
        with TextIOWrapper(
            fileobj, encoding="utf-8", newline="", write_through=True
        ) as buffer:
            yield buffer
            fileobj.seek(0)
            _put_object(bucket, key, fileobj)


def parse_url(url: str) -> S3Url:
    """Parse and validate url."""
    bucket, key = url.split("/", 1)
    if not bucket:
        raise ValueError("Invalid url: bucket is missing.")
    if not key:
        raise ValueError("Invalid url: key is missing.")
    return bucket, key


def write_csv_s3(
    url: str,
    queryset: QuerySet,
    *columns: str,
    header: bool = True,
    max_rows: int = MAX_ROWS,
) -> int:
    """Write a csv to S3."""
    bucket, key = parse_url(url)
    with s3_upload(bucket, key) as fileobj:
        return write_csv(fileobj, queryset, *columns, header=header, max_rows=max_rows)



from django.conf import settings

# Cap on the number of rows that can be downloaded
MAX_ROWS = getattr(settings, "CSV_DOWNLOAD_MAX_ROWS", 10000)

# Default page size used by PagedQuerySetWriter
DEFAULT_PAGE_SIZE = getattr(settings, "CSV_DOWNLOAD_PAGE_SIZE", 10000)



"""SFTP upload functions."""
import contextlib
import io
import tempfile
from typing import Generator, Optional, TextIO, Tuple
from urllib.parse import urlparse

from django.db.models import QuerySet

from .csv import write_csv
from .settings import MAX_ROWS

try:
    import paramiko
except ImportError:
    raise ImportError("You cannot use the django_csv.sftp module without paramiko.")

# type used to represent "username:password@hostname:port/path" when parsed
SFTPUrl = Tuple[str, str, str, int, str]


@contextlib.contextmanager
def sftp_upload(
    client: paramiko.SFTPClient, filepath: str
) -> Generator[TextIO, None, None]:
    """
    Return context manager that can be used to upload to SFTP.

    This context manager writes to a NamedTemporaryFile and then uses
    the paramiko client `putfo` method to upload the file.

    >>> with sftp_upload(client, "path/to/file.csv) as fileobj:  # noqa
    ...     write_csv(fileobj, queryset, "col1", "col2")

    The client is not created inside the generator as it would have to then
    handle all of the various authentication mechanisms that SSH / paramiko
    support, and that is really up to the calling code. There is a convenience
    function, `write_csv_sftp` that handles the basic use case of a username
    and password auth.

    """
    with tempfile.TemporaryFile() as fileobj:
        with io.TextIOWrapper(
            fileobj,
            encoding="utf-8",
            newline="",
            write_through=True,
        ) as buffer:
            yield buffer
            fileobj.seek(0)
            client.putfo(fileobj, filepath)


@contextlib.contextmanager
def sftp_client(
    hostname: str,
    username: str,
    port: int = 22,
    password: Optional[str] = None,
    pkey: Optional[paramiko.PKey] = None,
) -> Generator[paramiko.SFTPClient, None, None]:
    """
    Connect to SFTP server and return client object.

    This is done as a context manager so that we always close the connection
    after use.

    """
    if not any([pkey, password]):
        raise ValueError("Unable to connect via SFTP without pkey or password")
    transport = paramiko.Transport((hostname, port))
    transport.connect(
        username=username,
        password=password,
        pkey=pkey,
    )
    if client := paramiko.SFTPClient.from_transport(transport):
        yield client
        client.close()
    else:
        raise Exception("Unable to connect to remote server.")


def parse_url(url: str) -> SFTPUrl:
    """Parse and validate url."""
    if not url.startswith("sftp://"):
        url = "sftp://" + url
    parts = urlparse(url)
    if parts.scheme != "sftp":
        raise ValueError("Invalid url: scheme must be 'sftp'.")
    if not parts.hostname:
        raise ValueError("Invalid url: hostname is missing.")
    if not parts.username:
        raise ValueError("Invalid url: username is missing.")
    if not parts.password:
        raise ValueError("Invalid url: password is missing.")
    if not parts.port:
        raise ValueError("Invalid url: port is missing.")
    if not parts.path.lstrip("/"):
        raise ValueError("Invalid url: file path is missing.")
    return (
        parts.hostname,
        parts.username,
        parts.password,
        parts.port,
        parts.path.lstrip("/"),
    )


def write_csv_sftp(
    url: str,
    queryset: QuerySet,
    *columns: str,
    header: bool = True,
    max_rows: int = MAX_ROWS,
) -> int:
    """
    Write a csv to sftp.

    The url arg must be in the form:

        sftp://username:password@hostname:port/path

    All parts must exist, and a ValueError is raised if any
    are missing.

    """
    hostname, username, password, port, path = parse_url(url)
    with sftp_client(hostname, username, port=port, password=password) as client:
        with sftp_upload(client, path) as fileobj:
            return write_csv(
                fileobj, queryset, *columns, header=header, max_rows=max_rows
            )
